/* SPDX-License-Identifier: MIT OR X11
 *
 * Copyright © 2024 Enrico Weigelt, metux IT consult <info@metux.net>
 */
#ifndef _XSERVER_DRI_PRIV_H
#define _XSERVER_DRI_PRIV_H

#include "dri.h"

extern Bool DRIExtensionInit(void);

#endif /* _XSERVER_DRI_PRIV_H */
